
import MetalKit

enum ImageScalingMode {
    case aspectFit
    case aspectFill
    case scaleToFill
    case dontResize
}

class MetalImageView : MTKView {
    var metalContext: MetalContext
    var scalingMode: ImageScalingMode = .aspectFit
    var texture: MTLTexture?

    private var pipelineState: MTLRenderPipelineState!
    private var vertexBuffer: MTLBuffer!
    
    private var time: Float = 0
    
    override init(frame frameRect: CGRect, device: MTLDevice?) {
        metalContext = MetalContext(device: device ?? MTLCreateSystemDefaultDevice()!)
        super.init(frame: frameRect, device: device)
        self.device = device
        makeVertexBuffer()
        makeRenderPipelineState()
    }
    
    required init(coder: NSCoder) {
        metalContext = MetalContext.default
        super.init(coder: coder)
        self.device = metalContext.device
        makeVertexBuffer()
        makeRenderPipelineState()
    }

    func makeVertexBuffer() {
        // Vertex list for a full-screen quad described as a triangle strip.
        // Each vertex has two clip-space coordinates, followed by two texture coordinates.
        var vertices: [Float] = [
            -1, -1, 0, 1,  // lower left
            -1,  1, 0, 0,  // upper left
             1, -1, 1, 1,  // lower right
             1,  1, 1, 0,  // upper right
        ]
        
        vertexBuffer = metalContext.device.makeBuffer(bytes:&vertices, length:16 * MemoryLayout<Float>.size, options:[])
    }
    
    func makeRenderPipelineState() {
        let vertexDescriptor = MTLVertexDescriptor()
        vertexDescriptor.attributes[0].format = .float2 // position
        vertexDescriptor.attributes[0].offset = 0
        vertexDescriptor.attributes[0].bufferIndex = 0
        vertexDescriptor.attributes[1].format = .float2 // texCoords
        vertexDescriptor.attributes[1].offset = 2 * MemoryLayout<Float>.size
        vertexDescriptor.attributes[1].bufferIndex = 0
        vertexDescriptor.layouts[0].stride = 4 * MemoryLayout<Float>.size
        vertexDescriptor.layouts[0].stepFunction = .perVertex
        vertexDescriptor.layouts[0].stepRate = 1

        guard let library = metalContext.device.makeDefaultLibrary() else {
            print("Unable to locate default Metal library in main bundle")
            return
        }
        let vertexFunction = library.makeFunction(name: "vertex_reshape")
        let fragmentFunction = library.makeFunction(name: "fragment_texture")
        
        let pipelineDescriptor = MTLRenderPipelineDescriptor()
        pipelineDescriptor.vertexFunction = vertexFunction
        pipelineDescriptor.fragmentFunction = fragmentFunction
        pipelineDescriptor.vertexDescriptor = vertexDescriptor
        pipelineDescriptor.colorAttachments[0].pixelFormat = self.colorPixelFormat
        
        do {
            pipelineState = try metalContext.device.makeRenderPipelineState(descriptor: pipelineDescriptor)
        } catch {
            print("Error occurred when making render pipeline state: \(error)")
        }
    }

    var scaleMatrix: float2x2 {
        var matrix = matrix_identity_float2x2
        
        guard let texture = texture else { return matrix }
        
        let imageAspect = Float(texture.width) / Float(texture.height)
        let viewAspect = Float(drawableSize.width) / Float(drawableSize.height)

        switch scalingMode {
        case .scaleToFill:
            return matrix
        case .aspectFit:
            if (imageAspect < viewAspect) {
                matrix[0][0] = (imageAspect / viewAspect)
            } else {
                matrix[1][1] = (viewAspect / imageAspect)
            }
            return matrix
        case .aspectFill:
            if (imageAspect > viewAspect) {
                matrix[0][0] = (imageAspect / viewAspect)
            } else {
                matrix[1][1] = (viewAspect / imageAspect)
            }
            return matrix
        case .dontResize:
            matrix[0][0] = (Float(texture.width) / Float(drawableSize.width))
            matrix[1][1] = (Float(texture.height) / Float(drawableSize.height))
            return matrix
        }
    }

    override func draw(_ dirtyRect: CGRect) {
        guard let commandBuffer = metalContext.commandQueue.makeCommandBuffer() else { return }
        
        guard let passDescriptor = currentRenderPassDescriptor else { return }
        
        guard let renderCommandEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: passDescriptor) else { return }
        
        time += 1 / 60.0
        
        if let texture = texture {
            var scaleMatrix = self.scaleMatrix
            renderCommandEncoder.setVertexBytes(&scaleMatrix, length:4 * MemoryLayout<Float>.size, index:1)
            renderCommandEncoder.setVertexBuffer(vertexBuffer, offset:0, index:0)
            renderCommandEncoder.setFragmentBytes(&time, length: MemoryLayout<Float>.size, index: 0)
            renderCommandEncoder.setRenderPipelineState(pipelineState)
            renderCommandEncoder.setFragmentTexture(texture, index:0)
            renderCommandEncoder.drawPrimitives(type: .triangleStrip, vertexStart:0, vertexCount:4)
        }
        
        renderCommandEncoder.endEncoding()
        
        if let drawable = currentDrawable {
            commandBuffer.present(drawable)
        }
        
        commandBuffer.commit()
    }
}
